@@ -1,24 +1,36 @@
-/* Copyright (c) 2008, 2011, Oracle and/or its affiliates. All rights reserved.
+/* Copyright (c) 2008, 2021, Oracle and/or its affiliates.
 
    This program is free software; you can redistribute it and/or modify
-   it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; version 2 of the License.
+   it under the terms of the GNU General Public License, version 2.0,
+   as published by the Free Software Foundation.
+
+   This program is also distributed with certain software (including
+   but not limited to OpenSSL) that is licensed under separate terms,
+   as designated in a particular file or component or in included license
+   documentation.  The authors of MySQL hereby grant you an additional
+   permission to link the program and your derivative works with the
+   separately licensed software that they have included with MySQL.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-   GNU General Public License for more details.
+   GNU General Public License, version 2.0, for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA */
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA */
 
 #include "DblqhProxy.hpp"
 #include "Dblqh.hpp"
 #include "DblqhCommon.hpp"
 
+#include <signaldata/NodeRecoveryStatusRep.hpp>
 #include <signaldata/StartFragReq.hpp>
 #include <signaldata/ExecFragReq.hpp>
+#include <signaldata/DumpStateOrd.hpp>
+
+#define JAM_FILE_ID 442
+
 
 DblqhProxy::DblqhProxy(Block_context& ctx) :
   LocalProxy(DBLQH, ctx),
@@ -46,7 +58,7 @@ DblqhProxy::DblqhProxy(Block_context& ctx) :
   // GSN_LCP_FRAG_ORD
   addRecSignal(GSN_LCP_FRAG_ORD, &DblqhProxy::execLCP_FRAG_ORD);
   addRecSignal(GSN_LCP_FRAG_REP, &DblqhProxy::execLCP_FRAG_REP);
-  addRecSignal(GSN_END_LCP_CONF, &DblqhProxy::execEND_LCP_CONF);
+  addRecSignal(GSN_END_LCPCONF, &DblqhProxy::execEND_LCPCONF);
   addRecSignal(GSN_LCP_COMPLETE_REP, &DblqhProxy::execLCP_COMPLETE_REP);
 
   addRecSignal(GSN_EMPTY_LCP_REQ, &DblqhProxy::execEMPTY_LCP_REQ);
@@ -77,6 +89,8 @@ DblqhProxy::DblqhProxy(Block_context& ctx) :
   // GSN_START_RECREQ
   addRecSignal(GSN_START_RECREQ, &DblqhProxy::execSTART_RECREQ);
   addRecSignal(GSN_START_RECCONF, &DblqhProxy::execSTART_RECCONF);
+  addRecSignal(GSN_LOCAL_RECOVERY_COMP_REP,
+               &DblqhProxy::execLOCAL_RECOVERY_COMP_REP);
 
   // GSN_LQH_TRANSREQ
   addRecSignal(GSN_LQH_TRANSREQ, &DblqhProxy::execLQH_TRANSREQ);
@@ -364,6 +378,9 @@ DblqhProxy::execLQHFRAGREQ(Signal* signal)
   LqhFragReq* req = (LqhFragReq*)signal->getDataPtrSend();
   Uint32 instance = getInstanceKey(req->tableId, req->fragId);
 
+  /* Ensure instance hasn't quietly mapped back to proxy! */
+  ndbrequire(signal->getSendersBlockRef() != reference());
+
   // wl4391_todo impl. method that fakes senders block-ref
   sendSignal(numberToRef(DBLQH, instance, getOwnNodeId()),
              GSN_LQHFRAGREQ, signal, signal->getLength(), JBB);
@@ -468,9 +485,22 @@ DblqhProxy::execLCP_FRAG_ORD(Signal* signal)
   if (c_lcpRecord.m_state == LcpRecord::L_IDLE)
   {
     jam();
+    if (c_lcpRecord.m_lcpId == req->lcpId)
+    {
+      jam();
+      /**
+       * In master take over the new master doesn't get the perfect view
+       * of our state immediately. So he can resend LCP_FRAG_ORDs that
+       * were already sent. In this case we have already completed the
+       * LCP and the new master is sending messages which it doesn't
+       * know has been sent already.
+       */
+      D("LCP: Already completed" << V(req->lcpId));
+      return;
+    }
     D("LCP: start" << V(req->lcpId));
-    c_lcpRecord.m_state = LcpRecord::L_STARTING;
     c_lcpRecord.m_lcpId = req->lcpId;
+    c_lcpRecord.m_keepGci = req->keepGci;
     c_lcpRecord.m_lcp_frag_rep_cnt = 0;
     c_lcpRecord.m_lcp_frag_ord_cnt = 0;
     c_lcpRecord.m_complete_outstanding = 0;
@@ -491,9 +521,21 @@ DblqhProxy::execLCP_FRAG_ORD(Signal* signal)
 
   jam();
   D("LCP: continue" << V(req->lcpId) << V(c_lcpRecord.m_lcp_frag_ord_cnt));
-  ndbrequire(c_lcpRecord.m_state == LcpRecord::L_RUNNING);
   ndbrequire(c_lcpRecord.m_lcpId == req->lcpId);
 
+  if (c_lcpRecord.m_lastFragmentFlag)
+  {
+    jam();
+    /**
+     * We receive a new LCP_FRAG_ORD after already receiving a last flag on
+     * the LCP. We can safely drop this signal since it must have been
+     * generated by a new master that didn't know that the old master already
+     * sent the last fragment flag.
+     */
+    D("LCP: lastFragmentFlag already received");
+    return;
+  }
+  ndbrequire(c_lcpRecord.m_state == LcpRecord::L_RUNNING);
   if (lcp_complete_ord)
   {
     jam();
@@ -524,6 +566,7 @@ DblqhProxy::execLCP_FRAG_ORD(Signal* signal)
   {
     jam();
     /**
+     * Table has been dropped.
      * Send to lqh-0...that will handle it...
      */
     sendSignal(workerRef(0),
@@ -541,14 +584,28 @@ DblqhProxy::execLCP_FRAG_ORD(Signal* signal)
 void
 DblqhProxy::execLCP_FRAG_REP(Signal* signal)
 {
-  ndbrequire(signal->getLength() == LcpFragRep::SignalLength);
-
   LcpFragRep* conf = (LcpFragRep*)signal->getDataPtr();
-
   ndbrequire(c_lcpRecord.m_state == LcpRecord::L_RUNNING);
+  c_lcpRecord.m_lcp_frag_rep_cnt++;
+
+  if (signal->length() == 1)
+  {
+    /**
+     * An ignored signal is reported back to keep track of number of
+     * outstanding LCP_FRAG_ORD operations.
+     */
+    jam();
+    if (c_lcpRecord.m_lastFragmentFlag &&
+        (getNoOfOutstanding(c_lcpRecord) == 0))
+    {
+      jam();
+      completeLCP_1(signal);
+    }
+    return;
+  }
+  ndbrequire(signal->getLength() == LcpFragRep::SignalLength);
   ndbrequire(c_lcpRecord.m_lcpId == conf->lcpId);
 
-  c_lcpRecord.m_lcp_frag_rep_cnt++;
   D("LCP: rep" << V(conf->lcpId) << V(c_lcpRecord.m_lcp_frag_rep_cnt));
 
   /**
@@ -592,8 +649,12 @@ DblqhProxy::completeLCP_1(Signal* signal)
    *   they will reply with LCP_COMPLETE_REP
    */
   LcpFragOrd* ord = (LcpFragOrd*)signal->getDataPtrSend();
+  ord->tableId = RNIL;
+  ord->fragmentId = RNIL;
+  ord->lcpNo = RNIL;
   ord->lcpId = c_lcpRecord.m_lcpId;
   ord->lastFragmentFlag = true;
+  ord->keepGci = c_lcpRecord.m_keepGci;
   for (Uint32 i = 0; i<c_workers; i++)
   {
     jam();
@@ -603,8 +664,8 @@ DblqhProxy::completeLCP_1(Signal* signal)
   }
 
   /**
-   * send END_LCP_REQ to all pgman instances (except "extra" pgman)
-   *   they will reply with END_LCP_CONF
+   * send END_LCPREQ to all pgman instances (except "extra" pgman)
+   *   they will reply with END_LCPCONF
    */
   EndLcpReq* req = (EndLcpReq*)signal->getDataPtrSend();
   req->senderData= 0;
@@ -616,7 +677,7 @@ DblqhProxy::completeLCP_1(Signal* signal)
     jam();
     c_lcpRecord.m_complete_outstanding++;
     sendSignal(numberToRef(PGMAN, workerInstance(i), getOwnNodeId()),
-               GSN_END_LCP_REQ, signal, EndLcpReq::SignalLength, JBB);
+               GSN_END_LCPREQ, signal, EndLcpReq::SignalLength, JBA);
   }
 }
 
@@ -637,7 +698,7 @@ DblqhProxy::execLCP_COMPLETE_REP(Signal* signal)
 }
 
 void
-DblqhProxy::execEND_LCP_CONF(Signal* signal)
+DblqhProxy::execEND_LCPCONF(Signal* signal)
 {
   jamEntry();
   ndbrequire(c_lcpRecord.m_state == LcpRecord::L_COMPLETING_1 ||
@@ -692,7 +753,7 @@ DblqhProxy::completeLCP_2(Signal* signal)
   // NOTE: ugly to use MaxLqhWorkers directly
   Uint32 instance = c_workers + 1;
   sendSignal(numberToRef(PGMAN, instance, getOwnNodeId()),
-             GSN_END_LCP_REQ, signal, EndLcpReq::SignalLength, JBB);
+             GSN_END_LCPREQ, signal, EndLcpReq::SignalLength, JBA);
 }
 
 
@@ -714,15 +775,15 @@ DblqhProxy::completeLCP_3(Signal* signal)
   req->backupId= c_lcpRecord.m_lcpId;
 
   // no reply from this
-  sendSignal(TSMAN_REF, GSN_END_LCP_REQ, signal,
-             EndLcpReq::SignalLength, JBB);
+  sendSignal(TSMAN_REF, GSN_END_LCPREQ, signal,
+             EndLcpReq::SignalLength, JBA);
 
   if (c_lcpRecord.m_lcp_frag_rep_cnt)
   {
     jam();
     c_lcpRecord.m_complete_outstanding++;
-    sendSignal(LGMAN_REF, GSN_END_LCP_REQ, signal,
-               EndLcpReq::SignalLength, JBB);
+    sendSignal(LGMAN_REF, GSN_END_LCPREQ, signal,
+               EndLcpReq::SignalLength, JBA);
   }
   else
   {
@@ -779,9 +840,6 @@ DblqhProxy::checkSendEMPTY_LCP_CONF_impl(Signal* signal)
     jam();
     conf->idle = true;
     break;
-  case LcpRecord::L_STARTING:
-    jam();
-    return;
   case LcpRecord::L_RUNNING:{
     jam();
     if (getNoOfOutstanding(c_lcpRecord) == 0)
@@ -1180,6 +1238,10 @@ DblqhProxy::execSTART_RECREQ(Signal* signal)
   const StartRecReq* req = (const StartRecReq*)signal->getDataPtr();
   Ss_START_RECREQ& ss = ssSeize<Ss_START_RECREQ>();
   ss.m_req = *req;
+  ss.restoreFragCompletedCount = 0;
+  ss.undoDDCompletedCount = 0;
+  ss.execREDOLogCompletedCount = 0;
+  ss.phaseToSend = 0;
 
   // seize records for sub-ops
   Uint32 i;
@@ -1198,6 +1260,72 @@ DblqhProxy::execSTART_RECREQ(Signal* signal)
   sendREQ(signal, ss);
 }
 
+void
+DblqhProxy::execLOCAL_RECOVERY_COMP_REP(Signal *signal)
+{
+  LocalRecoveryCompleteRep *rep =
+    (LocalRecoveryCompleteRep*) &signal->theData[0];
+
+  LocalRecoveryCompleteRep::PhaseIds phaseId =
+    (LocalRecoveryCompleteRep::PhaseIds)rep->phaseId;
+
+  Uint32 ssId = rep->senderData;
+  Ss_START_RECREQ& ss = ssFind<Ss_START_RECREQ>(ssId);
+
+  switch (phaseId)
+  {
+  case LocalRecoveryCompleteRep::RESTORE_FRAG_COMPLETED:
+  {
+    ss.restoreFragCompletedCount++;
+    if (ss.restoreFragCompletedCount < c_workers)
+    {
+      jam();
+      return;
+    }
+    break;
+  }
+  case LocalRecoveryCompleteRep::UNDO_DD_COMPLETED:
+  {
+    ss.undoDDCompletedCount++;
+    if (ss.undoDDCompletedCount < c_workers)
+    {
+      jam();
+      return;
+    }
+    break;
+  }
+  case LocalRecoveryCompleteRep::EXECUTE_REDO_LOG_COMPLETED:
+  {
+    ss.execREDOLogCompletedCount++;
+    if (ss.execREDOLogCompletedCount < c_workers)
+    {
+      jam();
+      return;
+    }
+    break;
+  }
+  default:
+    ndbrequire(false);
+  }
+  /* All LDM workers have completed this phase */
+  ndbrequire(Uint32(ss.phaseToSend) == Uint32(phaseId));
+  ss.phaseToSend++;
+  Uint32 masterNodeId = refToNode(ss.m_req.senderRef);
+  Uint32 master_version = getNodeInfo(masterNodeId).m_version;
+  if (master_version >= NDBD_NODE_RECOVERY_STATUS_VERSION)
+  {
+    /* Only send this information to masters that have code to handle it. */
+    jam();
+    rep->nodeId = getOwnNodeId();
+    rep->phaseId = (Uint32)phaseId;
+    sendSignal(ss.m_req.senderRef, GSN_LOCAL_RECOVERY_COMP_REP, signal,
+               LocalRecoveryCompleteRep::SignalLengthMaster, JBB);
+    return;
+  }
+  jam();
+  return;
+}
+
 void
 DblqhProxy::sendSTART_RECREQ(Signal* signal, Uint32 ssId, SectionHandle* handle)
 {
@@ -1242,9 +1370,11 @@ DblqhProxy::sendSTART_RECCONF(Signal* signal, Uint32 ssId)
     /**
      * There should be no disk-ops in flight here...check it
      */
-    signal->theData[0] = 12003;
+    signal->theData[0] = DumpStateOrd::LgmanCheckCallbacksClear;
     sendSignal(LGMAN_REF, GSN_DUMP_STATE_ORD, signal, 1, JBB);
 
+    ndbrequire(ss.phaseToSend ==
+          Uint32(LocalRecoveryCompleteRep::LOCAL_RECOVERY_COMPLETED));
     StartRecConf* conf = (StartRecConf*)signal->getDataPtrSend();
     conf->startingNodeId = getOwnNodeId();
     conf->senderData = ss.m_req.senderData;
@@ -1362,8 +1492,17 @@ DblqhProxy::execLQH_TRANSREQ(Signal* signal)
   }
   const LqhTransReq* req = (const LqhTransReq*)signal->getDataPtr();
   Ss_LQH_TRANSREQ& ss = ssSeize<Ss_LQH_TRANSREQ>();
+  ss.m_maxInstanceId = 0;
   ss.m_req = *req;
-  ndbrequire(signal->getLength() == LqhTransReq::SignalLength);
+  if (signal->getLength() < LqhTransReq::SignalLength)
+  {
+    /**
+     * TC that performs take over doesn't suppport taking over one
+     * TC instance at a time
+     */
+     ss.m_req.instanceId = RNIL;
+  }
+  ndbrequire(signal->getLength() <= LqhTransReq::SignalLength);
   sendREQ(signal, ss);
 
   /**
@@ -1473,8 +1612,25 @@ DblqhProxy::sendLQH_TRANSCONF(Signal* signal, Uint32 ssId)
 {
   Ss_LQH_TRANSREQ& ss = ssFind<Ss_LQH_TRANSREQ>(ssId);
 
-  if (ss.m_conf.operationStatus != LqhTransConf::LastTransConf) {
+  if (ss.m_conf.operationStatus == LqhTransConf::LastTransConf) 
+  {
     jam();
+
+    /**
+     * Maintain running max instance id, based on max seen 
+     * in this pass, which is only sent in LastTransConf 
+     * LQH_TRANSREQ from each instance.
+     */
+    if (ss.m_conf.maxInstanceId > ss.m_maxInstanceId)
+    {
+      ndbrequire(ss.m_conf.maxInstanceId < NDBMT_MAX_BLOCK_INSTANCES);
+      ss.m_maxInstanceId = ss.m_conf.maxInstanceId;
+    }
+  }
+  else
+  {
+    jam();
+    /* Forward conf to the requesting TC, and wait for more */
     LqhTransConf* conf = (LqhTransConf*)signal->getDataPtrSend();
     *conf = ss.m_conf;
     conf->tcRef = ss.m_req.senderData;
@@ -1494,6 +1650,7 @@ DblqhProxy::sendLQH_TRANSCONF(Signal* signal, Uint32 ssId)
     conf->tcRef = ss.m_req.senderData;
     conf->lqhNodeId = getOwnNodeId();
     conf->operationStatus = LqhTransConf::LastTransConf;
+    conf->maxInstanceId = ss.m_maxInstanceId;
     sendSignal(ss.m_req.senderRef, GSN_LQH_TRANSCONF,
                signal, LqhTransConf::SignalLength, JBB);
   } else {
