@@ -1,5 +1,5 @@
 /*
-   Copyright (c) 2005, 2017, Oracle and/or its affiliates. All rights reserved.
+   Copyright (c) 2005, 2021, Oracle and/or its affiliates.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License, version 2.0,
@@ -121,9 +121,6 @@ static struct my_option my_long_options[] =
   { "type", NDB_OPT_NOSHORT, "Type of node/connection",
     (uchar**) &g_type, (uchar**) &g_type,
     0, GET_STR, REQUIRED_ARG, 0, 0, 0, 0, 0, 0},
-  { "id", NDB_OPT_NOSHORT, "Nodeid",
-    (uchar**) &g_nodeid, (uchar**) &g_nodeid,
-    0, GET_INT, REQUIRED_ARG, 0, 0, 0, 0, 0, 0},
   { "nodeid", NDB_OPT_NOSHORT, "Nodeid",
     (uchar**) &g_nodeid, (uchar**) &g_nodeid,
     0, GET_INT, REQUIRED_ARG, 0, 0, 0, 0, 0, 0},
@@ -352,6 +349,7 @@ main(int argc, char** argv){
     }
   }
   printf("\n");
+  ndb_free_defaults(argv);
   return 0;
 }
 
@@ -496,18 +494,17 @@ helper(Vector<Apply*>& select, const char * str)
       select.push_back(new ParamApply(CFG_NODE_HOST, "host"));
       select.push_back(new NodeTypeApply("type"));
     }
-    else if (strcasecmp(str, "nodeid") == 0 ||
-             strcasecmp(str, "id") == 0)
+    else if (native_strcasecmp(str, "nodeid") == 0)
     {
       select.push_back(new ParamApply(CFG_NODE_ID, "nodeid"));
       retflag = true;
     }
-    else if (strncasecmp(str, "host", 4) == 0)
+    else if (native_strncasecmp(str, "host", 4) == 0)
     {
       select.push_back(new ParamApply(CFG_NODE_HOST, "host"));
       retflag = true;
     }
-    else if (strcasecmp(str, "type") == 0)
+    else if (native_strcasecmp(str, "type") == 0)
     {
       select.push_back(new NodeTypeApply("type"));
       retflag = true;
@@ -515,7 +512,7 @@ helper(Vector<Apply*>& select, const char * str)
   }
   else if (g_section == CFG_SECTION_CONNECTION)
   {
-    if (all || strcasecmp(str, "type") == 0)
+    if (all || native_strcasecmp(str, "type") == 0)
     {
       select.push_back(new ConnectionTypeApply("type"));
       retflag = true;
@@ -541,7 +538,7 @@ helper(Vector<Apply*>& select, const char * str)
         ||
         (g_section == CFG_SECTION_SYSTEM))
       {
-        if (all || strcasecmp(ConfigInfo::m_ParamInfo[p]._fname, str) == 0)
+        if (all || native_strcasecmp(ConfigInfo::m_ParamInfo[p]._fname, str) == 0)
         {
           select.push_back(new ParamApply(ConfigInfo::m_ParamInfo[p]._paramId,
             ConfigInfo::m_ParamInfo[p]._fname));
@@ -658,7 +655,7 @@ Match::eval(const Iter& iter)
   } 
   else if(iter.get(m_key, &val64) == 0)
   {
-    if(strtoll(m_value.c_str(), (char **)NULL, 10) != (long long)val64)
+    if(my_strtoll(m_value.c_str(), (char **)NULL, 10) != (long long)val64)
       return 0;
   }
   else if(iter.get(m_key, &valc) == 0)
@@ -680,36 +677,47 @@ HostMatch::eval(const Iter& iter)
   
   if(iter.get(m_key, &valc) == 0)
   {
-	  struct hostent *h1, *h2, copy1;
-	  char *addr1;
+    struct hostent *h1, *h2, copy1;
+    char *addr1;
 
-	  h1 = gethostbyname(m_value.c_str());
-	  if (h1 == NULL) {
-		  return 0;
-	  }
+    if (m_value.empty())
+    {
+      return 0;
+    }
+    h1 = gethostbyname(m_value.c_str());
+    if (h1 == NULL)
+    {
+      return 0;
+    }
 
-	  // gethostbyname returns a pointer to a static structure
-	  // so we need to copy the results before doing the next call
-	  memcpy(&copy1, h1, sizeof(struct hostent));
-	  addr1 = (char *)malloc(copy1.h_length);
-	  NdbAutoPtr<char> tmp_aptr(addr1);
-	  memcpy(addr1, h1->h_addr, copy1.h_length);
+    // gethostbyname returns a pointer to a static structure
+    // so we need to copy the results before doing the next call
+    memcpy(&copy1, h1, sizeof(struct hostent));
+    addr1 = (char *)malloc(copy1.h_length);
+    NdbAutoPtr<char> tmp_aptr(addr1);
+    memcpy(addr1, h1->h_addr, copy1.h_length);
 
-	  h2 = gethostbyname(valc);
-	  if (h2 == NULL) {
-		  return 0;
-	  }
+    if (valc == NULL || strlen(valc) == 0)
+    {
+      return 0;
+    }
+    h2 = gethostbyname(valc);
+    if (h2 == NULL)
+    {
+      return 0;
+    }
 
-	  if (copy1.h_addrtype != h2->h_addrtype) {
-		  return 0;
-	  }
+    if (copy1.h_addrtype != h2->h_addrtype)
+    {
+      return 0;
+    }
 
-	  if (copy1.h_length != h2->h_length) 
-	  {
-		  return 0;
-	  }
-	  
-	  return 0 ==  memcmp(addr1, h2->h_addr, copy1.h_length);	  
+    if (copy1.h_length != h2->h_length) 
+    {
+      return 0;
+    }
+ 
+    return 0 ==  memcmp(addr1, h2->h_addr, copy1.h_length);	  
   }
 
   return 0;
@@ -794,7 +802,7 @@ fetch_configuration(int from_node)
     goto noconnect;
   }
 
-  if(ndb_mgm_connect(mgm, opt_connect_retries, opt_connect_retry_delay, 1))
+  if(ndb_mgm_connect(mgm, opt_connect_retries - 1, opt_connect_retry_delay, 1))
   {
     fprintf(stderr, "Connect failed");
     fprintf(stderr, " code: %d, msg: %s\n",
